# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class CreateCloudResource(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'aws_vpc_id': 'str',
        'aws_subnet_ids_with_availability_zones': 'list[SubnetIdWithAvailabilityZoneAWS]',
        'aws_iam_role_arns': 'list[str]',
        'aws_security_groups': 'list[str]',
        'aws_s3_id': 'str',
        'aws_efs_id': 'str',
        'aws_efs_mount_target_ip': 'str',
        'aws_cloudformation_stack_id': 'str',
        'gcp_vpc_id': 'str',
        'gcp_subnet_ids': 'list[str]'
    }

    attribute_map = {
        'aws_vpc_id': 'aws_vpc_id',
        'aws_subnet_ids_with_availability_zones': 'aws_subnet_ids_with_availability_zones',
        'aws_iam_role_arns': 'aws_iam_role_arns',
        'aws_security_groups': 'aws_security_groups',
        'aws_s3_id': 'aws_s3_id',
        'aws_efs_id': 'aws_efs_id',
        'aws_efs_mount_target_ip': 'aws_efs_mount_target_ip',
        'aws_cloudformation_stack_id': 'aws_cloudformation_stack_id',
        'gcp_vpc_id': 'gcp_vpc_id',
        'gcp_subnet_ids': 'gcp_subnet_ids'
    }

    def __init__(self, aws_vpc_id=None, aws_subnet_ids_with_availability_zones=None, aws_iam_role_arns=None, aws_security_groups=None, aws_s3_id=None, aws_efs_id=None, aws_efs_mount_target_ip=None, aws_cloudformation_stack_id=None, gcp_vpc_id=None, gcp_subnet_ids=None, local_vars_configuration=None):  # noqa: E501
        """CreateCloudResource - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._aws_vpc_id = None
        self._aws_subnet_ids_with_availability_zones = None
        self._aws_iam_role_arns = None
        self._aws_security_groups = None
        self._aws_s3_id = None
        self._aws_efs_id = None
        self._aws_efs_mount_target_ip = None
        self._aws_cloudformation_stack_id = None
        self._gcp_vpc_id = None
        self._gcp_subnet_ids = None
        self.discriminator = None

        if aws_vpc_id is not None:
            self.aws_vpc_id = aws_vpc_id
        if aws_subnet_ids_with_availability_zones is not None:
            self.aws_subnet_ids_with_availability_zones = aws_subnet_ids_with_availability_zones
        if aws_iam_role_arns is not None:
            self.aws_iam_role_arns = aws_iam_role_arns
        if aws_security_groups is not None:
            self.aws_security_groups = aws_security_groups
        if aws_s3_id is not None:
            self.aws_s3_id = aws_s3_id
        if aws_efs_id is not None:
            self.aws_efs_id = aws_efs_id
        if aws_efs_mount_target_ip is not None:
            self.aws_efs_mount_target_ip = aws_efs_mount_target_ip
        if aws_cloudformation_stack_id is not None:
            self.aws_cloudformation_stack_id = aws_cloudformation_stack_id
        if gcp_vpc_id is not None:
            self.gcp_vpc_id = gcp_vpc_id
        if gcp_subnet_ids is not None:
            self.gcp_subnet_ids = gcp_subnet_ids

    @property
    def aws_vpc_id(self):
        """Gets the aws_vpc_id of this CreateCloudResource.  # noqa: E501

        The AWS vpc id of this cloud.  # noqa: E501

        :return: The aws_vpc_id of this CreateCloudResource.  # noqa: E501
        :rtype: str
        """
        return self._aws_vpc_id

    @aws_vpc_id.setter
    def aws_vpc_id(self, aws_vpc_id):
        """Sets the aws_vpc_id of this CreateCloudResource.

        The AWS vpc id of this cloud.  # noqa: E501

        :param aws_vpc_id: The aws_vpc_id of this CreateCloudResource.  # noqa: E501
        :type: str
        """

        self._aws_vpc_id = aws_vpc_id

    @property
    def aws_subnet_ids_with_availability_zones(self):
        """Gets the aws_subnet_ids_with_availability_zones of this CreateCloudResource.  # noqa: E501

        A list of AWS subnet IDs with its availability zone of this cloud.  # noqa: E501

        :return: The aws_subnet_ids_with_availability_zones of this CreateCloudResource.  # noqa: E501
        :rtype: list[SubnetIdWithAvailabilityZoneAWS]
        """
        return self._aws_subnet_ids_with_availability_zones

    @aws_subnet_ids_with_availability_zones.setter
    def aws_subnet_ids_with_availability_zones(self, aws_subnet_ids_with_availability_zones):
        """Sets the aws_subnet_ids_with_availability_zones of this CreateCloudResource.

        A list of AWS subnet IDs with its availability zone of this cloud.  # noqa: E501

        :param aws_subnet_ids_with_availability_zones: The aws_subnet_ids_with_availability_zones of this CreateCloudResource.  # noqa: E501
        :type: list[SubnetIdWithAvailabilityZoneAWS]
        """

        self._aws_subnet_ids_with_availability_zones = aws_subnet_ids_with_availability_zones

    @property
    def aws_iam_role_arns(self):
        """Gets the aws_iam_role_arns of this CreateCloudResource.  # noqa: E501

        A list of 2 AWS IAM role arns of this cloud (the IAM role Anyscale uses to access this cloud & the IAM role assigned to nodes in the cluster).  # noqa: E501

        :return: The aws_iam_role_arns of this CreateCloudResource.  # noqa: E501
        :rtype: list[str]
        """
        return self._aws_iam_role_arns

    @aws_iam_role_arns.setter
    def aws_iam_role_arns(self, aws_iam_role_arns):
        """Sets the aws_iam_role_arns of this CreateCloudResource.

        A list of 2 AWS IAM role arns of this cloud (the IAM role Anyscale uses to access this cloud & the IAM role assigned to nodes in the cluster).  # noqa: E501

        :param aws_iam_role_arns: The aws_iam_role_arns of this CreateCloudResource.  # noqa: E501
        :type: list[str]
        """

        self._aws_iam_role_arns = aws_iam_role_arns

    @property
    def aws_security_groups(self):
        """Gets the aws_security_groups of this CreateCloudResource.  # noqa: E501

        A list of AWS security groups of this cloud.  # noqa: E501

        :return: The aws_security_groups of this CreateCloudResource.  # noqa: E501
        :rtype: list[str]
        """
        return self._aws_security_groups

    @aws_security_groups.setter
    def aws_security_groups(self, aws_security_groups):
        """Sets the aws_security_groups of this CreateCloudResource.

        A list of AWS security groups of this cloud.  # noqa: E501

        :param aws_security_groups: The aws_security_groups of this CreateCloudResource.  # noqa: E501
        :type: list[str]
        """

        self._aws_security_groups = aws_security_groups

    @property
    def aws_s3_id(self):
        """Gets the aws_s3_id of this CreateCloudResource.  # noqa: E501

        The AWS S3 id of this cloud.  # noqa: E501

        :return: The aws_s3_id of this CreateCloudResource.  # noqa: E501
        :rtype: str
        """
        return self._aws_s3_id

    @aws_s3_id.setter
    def aws_s3_id(self, aws_s3_id):
        """Sets the aws_s3_id of this CreateCloudResource.

        The AWS S3 id of this cloud.  # noqa: E501

        :param aws_s3_id: The aws_s3_id of this CreateCloudResource.  # noqa: E501
        :type: str
        """

        self._aws_s3_id = aws_s3_id

    @property
    def aws_efs_id(self):
        """Gets the aws_efs_id of this CreateCloudResource.  # noqa: E501

        The AWS EFS id of this cloud.  # noqa: E501

        :return: The aws_efs_id of this CreateCloudResource.  # noqa: E501
        :rtype: str
        """
        return self._aws_efs_id

    @aws_efs_id.setter
    def aws_efs_id(self, aws_efs_id):
        """Sets the aws_efs_id of this CreateCloudResource.

        The AWS EFS id of this cloud.  # noqa: E501

        :param aws_efs_id: The aws_efs_id of this CreateCloudResource.  # noqa: E501
        :type: str
        """

        self._aws_efs_id = aws_efs_id

    @property
    def aws_efs_mount_target_ip(self):
        """Gets the aws_efs_mount_target_ip of this CreateCloudResource.  # noqa: E501

        The AWS EFS mount target IP of this cloud.  # noqa: E501

        :return: The aws_efs_mount_target_ip of this CreateCloudResource.  # noqa: E501
        :rtype: str
        """
        return self._aws_efs_mount_target_ip

    @aws_efs_mount_target_ip.setter
    def aws_efs_mount_target_ip(self, aws_efs_mount_target_ip):
        """Sets the aws_efs_mount_target_ip of this CreateCloudResource.

        The AWS EFS mount target IP of this cloud.  # noqa: E501

        :param aws_efs_mount_target_ip: The aws_efs_mount_target_ip of this CreateCloudResource.  # noqa: E501
        :type: str
        """

        self._aws_efs_mount_target_ip = aws_efs_mount_target_ip

    @property
    def aws_cloudformation_stack_id(self):
        """Gets the aws_cloudformation_stack_id of this CreateCloudResource.  # noqa: E501

        The AWS CloudFormation stack id of this cloud.  # noqa: E501

        :return: The aws_cloudformation_stack_id of this CreateCloudResource.  # noqa: E501
        :rtype: str
        """
        return self._aws_cloudformation_stack_id

    @aws_cloudformation_stack_id.setter
    def aws_cloudformation_stack_id(self, aws_cloudformation_stack_id):
        """Sets the aws_cloudformation_stack_id of this CreateCloudResource.

        The AWS CloudFormation stack id of this cloud.  # noqa: E501

        :param aws_cloudformation_stack_id: The aws_cloudformation_stack_id of this CreateCloudResource.  # noqa: E501
        :type: str
        """

        self._aws_cloudformation_stack_id = aws_cloudformation_stack_id

    @property
    def gcp_vpc_id(self):
        """Gets the gcp_vpc_id of this CreateCloudResource.  # noqa: E501

        The GCP vpc id of this cloud.  # noqa: E501

        :return: The gcp_vpc_id of this CreateCloudResource.  # noqa: E501
        :rtype: str
        """
        return self._gcp_vpc_id

    @gcp_vpc_id.setter
    def gcp_vpc_id(self, gcp_vpc_id):
        """Sets the gcp_vpc_id of this CreateCloudResource.

        The GCP vpc id of this cloud.  # noqa: E501

        :param gcp_vpc_id: The gcp_vpc_id of this CreateCloudResource.  # noqa: E501
        :type: str
        """

        self._gcp_vpc_id = gcp_vpc_id

    @property
    def gcp_subnet_ids(self):
        """Gets the gcp_subnet_ids of this CreateCloudResource.  # noqa: E501

        A list of GCP subnet IDs of this cloud.  # noqa: E501

        :return: The gcp_subnet_ids of this CreateCloudResource.  # noqa: E501
        :rtype: list[str]
        """
        return self._gcp_subnet_ids

    @gcp_subnet_ids.setter
    def gcp_subnet_ids(self, gcp_subnet_ids):
        """Sets the gcp_subnet_ids of this CreateCloudResource.

        A list of GCP subnet IDs of this cloud.  # noqa: E501

        :param gcp_subnet_ids: The gcp_subnet_ids of this CreateCloudResource.  # noqa: E501
        :type: list[str]
        """

        self._gcp_subnet_ids = gcp_subnet_ids

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, CreateCloudResource):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, CreateCloudResource):
            return True

        return self.to_dict() != other.to_dict()



import os
from typing import Any, Dict, List, Optional, Union

import click
import tabulate
import yaml

from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client import ComputeTemplate, ProductionJob
from anyscale.client.openapi_client.models.decorated_list_service_api_model import (
    DecoratedListServiceAPIModel,
)
from anyscale.client.openapi_client.models.decorated_production_job import (
    DecoratedProductionJob,
)
from anyscale.client.openapi_client.models.decorated_production_service_v2_api_model import (
    DecoratedProductionServiceV2APIModel,
)
from anyscale.client.openapi_client.models.ha_job_states import HaJobStates
from anyscale.client.openapi_client.models.service_event_current_state import (
    ServiceEventCurrentState,
)
from anyscale.client.openapi_client.models.service_goal_states import ServiceGoalStates
from anyscale.controllers.base_controller import BaseController
from anyscale.controllers.job_controller import JobController
from anyscale.formatters.service_formatter import (
    format_service_config,
    format_service_config_v2,
)
from anyscale.models.service_model import ServiceConfig
from anyscale.project import infer_project_id
from anyscale.tables import ServicesTable
from anyscale.util import get_endpoint, is_anyscale_workspace
from anyscale.utils.runtime_env import (
    is_dir_remote_uri,
    override_runtime_env_for_local_working_dir,
    upload_and_rewrite_working_dir,
    upload_and_rewrite_workspace_working_dir,
)


def _is_v2_service_id(service_id: str) -> bool:
    return service_id.startswith("service2")


class ServiceController(BaseController):
    def __init__(
        self, log: Optional[BlockLogger] = None, initialize_auth_api_client: bool = True
    ):
        if log is None:
            log = BlockLogger()

        super().__init__(initialize_auth_api_client=initialize_auth_api_client)
        self.log = log
        self.job_controller = JobController(
            initialize_auth_api_client=initialize_auth_api_client
        )

    def _get_services_by_name(
        self,
        *,
        name: Optional[str] = None,
        project_id: Optional[str] = None,
        created_by_me: bool,
        max_items: int,
    ) -> List[DecoratedListServiceAPIModel]:
        """Makes an API call to get all services matching the provided filters.

        This returns both v1 and v2 services.
        """
        creator_id = None
        if created_by_me:
            user_info_response = self.api_client.get_user_info_api_v2_userinfo_get()
            creator_id = user_info_response.result.id

        services_list = []
        resp = self.api_client.list_services_api_v2_services_v2_get(
            project_id=project_id, name=name, creator_id=creator_id, count=10,
        )
        services_list.extend(resp.results)
        paging_token = resp.metadata.next_paging_token
        while paging_token is not None and len(services_list) < max_items:
            resp = self.api_client.list_services_api_v2_services_v2_get(
                project_id=project_id,
                name=name,
                creator_id=creator_id,
                count=10,
                paging_token=paging_token,
            )
            services_list.extend(resp.results)
            paging_token = resp.metadata.next_paging_token

        return services_list[:max_items]

    def _get_service_id_from_name(
        self, service_name: str, project_id: Optional[str]
    ) -> str:
        """Get the ID for a service by name.

        If project_id is specified, filter to that project, else don't filter on project_id and
        instead error if there are multiple services with the name.

        Raises an exception if there are zero or multiple services with the given name.
        """
        results = self._get_services_by_name(
            name=service_name, project_id=project_id, created_by_me=False, max_items=10
        )

        if len(results) == 0:
            raise click.ClickException(
                f"No service with name '{service_name}' was found. "
                "Please verify that this service exists and you have access to it."
            )
        elif len(results) > 1:
            raise click.ClickException(
                f"There are multiple services with name '{service_name}'. "
                "Please filter using --project-id or specify the --service-id instead. "
                f"Services found: \n{ServicesTable(results)}"
            )

        return results[0].id

    def get_service_id(
        self,
        *,
        service_id: Optional[str] = None,
        service_name: Optional[str] = None,
        service_config_file: Optional[str] = None,
        project_id: Optional[str] = None,
    ) -> str:
        """Get the service ID given the ID, name, or config file.

        This is a utility used by multiple CLI commands to standardize mapping these options
        to a service_id.

        The precedence is: service_id > service_name > service_config_file.

        If the project_id is specified directly or via the service_config_file it will be used
        to filter the query. Else we will try to find the service across all projects and error
        if there are multiple.
        """
        if service_id is not None:
            if service_name is not None or service_config_file is not None:
                raise click.ClickException(
                    "Only one of service ID, name, or config file should be specified."
                )
        elif service_name is not None:
            if service_config_file is not None:
                raise click.ClickException(
                    "Only one of service ID, name, or config file should be specified."
                )
            service_id = self._get_service_id_from_name(service_name, project_id)
        elif service_config_file is not None:
            service_config: ServiceConfig = self._read_service_config_file(
                service_config_file
            )
            # Allow the passed project_id to override the one in the file.
            project_id = project_id or service_config.project_id
            service_id = self._get_service_id_from_name(service_config.name, project_id)
        else:
            raise click.ClickException(
                "Service ID, name, or config file must be specified."
            )

        return service_id

    def deploy(
        self,
        service_config_file: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
        healthcheck_url: Optional[str] = None,
        is_entrypoint_cmd: Optional[bool] = False,
        entrypoint: Optional[List[str]] = None,
    ) -> None:
        entrypoint = entrypoint or []
        if is_anyscale_workspace() and is_entrypoint_cmd:
            # If is_entrypoint_cmd is true, service_config_file is expected to be "python"
            # Therefore, we need to combine args to create the final entrypoint
            entrypoint = [service_config_file, *entrypoint]
            config = self.generate_config_from_entrypoint(
                entrypoint,
                name=name,
                description=description,
                healthcheck_url=healthcheck_url,
            )
            project_id = infer_project_id(
                self.anyscale_api_client,
                self.api_client,
                self.log,
                project_id=config.project_id,
                cluster_compute_id=config.compute_config_id,
                cluster_compute=config.compute_config,
                cloud=config.cloud,
            )

            config.runtime_env = override_runtime_env_for_local_working_dir(
                config.runtime_env
            )
            config.project_id = project_id

            self.deploy_from_config(config)
        elif len(entrypoint) == 0:
            # Assume that service_config_file is a file and submit it.
            config = self.generate_config_from_file(
                service_config_file,
                name=name,
                description=description,
                healthcheck_url=healthcheck_url,
            )

            project_id = infer_project_id(
                self.anyscale_api_client,
                self.api_client,
                self.log,
                project_id=config.project_id,
                cluster_compute_id=config.compute_config_id,
                cluster_compute=config.compute_config,
                cloud=config.cloud,
            )

            config.project_id = project_id
            self.deploy_from_config(config)
        elif len(entrypoint) != 0:
            msg = (
                "Within an Anyscale Workspace, `anyscale service deploy` takes either a file, or a command. To submit a command, use `anyscale service deploy -- my command`."
                if is_anyscale_workspace()
                else "`anyscale service deploy` takes one argument, a YAML file configuration. Please use `anyscale service deploy my_file`."
            )
            raise click.ClickException(msg)

    def generate_config_from_entrypoint(
        self,
        entrypoint: List[str],
        name: Optional[str] = None,
        description: Optional[str] = None,
        healthcheck_url: Optional[str] = "/healthcheck",
    ) -> ServiceConfig:
        config_dict = {
            "entrypoint": " ".join(entrypoint),
            "name": name,
            "description": description,
            "healthcheck_url": healthcheck_url,
        }
        return self._populate_service_config(config_dict)

    def generate_config_from_file(
        self,
        service_config_file,
        *,
        name: Optional[str] = None,
        description: Optional[str] = None,
        healthcheck_url: Optional[str] = None,
        version: Optional[str] = None,
        canary_percent: Optional[int] = None,
        rollout_strategy: Optional[str] = None,
    ) -> ServiceConfig:
        """
        Given the input file path and user overrides, this method
        constructs the expected service configuration.

        Note that if the command is run from a workspace, some values
        may be overwritten.
        """
        service_config: ServiceConfig = self._read_service_config_file(
            service_config_file
        )
        if name:
            service_config.name = name

        if description:
            service_config.description = description

        if healthcheck_url:
            service_config.healthcheck_url = healthcheck_url

        if version:
            service_config.version = version

        if canary_percent is not None:
            service_config.canary_percent = canary_percent

        if rollout_strategy:
            service_config.rollout_strategy = rollout_strategy

        if service_config.rollout_strategy:
            service_config.rollout_strategy = service_config.rollout_strategy.upper()

        self._override_configs_in_workspaces(service_config)

        if service_config.runtime_env:
            # Service v1 specific override since Service v2's runtime_env
            # is under service_config.ray_serve_config["runtime_env"]
            service_config.runtime_env = upload_and_rewrite_working_dir(
                service_config.runtime_env
            )

        return service_config

    def _override_configs_in_workspaces(self, service_config: ServiceConfig) -> None:
        """
        We override some service configs if launched in workspaces.

        Please note that this method mutates service_config in-place.
        """
        if (
            is_anyscale_workspace()
            and "ANYSCALE_SESSION_ID" in os.environ
            and service_config.ray_serve_config
        ):
            existing_runtime_env = service_config.ray_serve_config.get(
                "runtime_env", {}
            )

            # If deploying service v2 from workspaces and no upload_path is specified,
            # we back up the current workspace content into S3
            working_dir = existing_runtime_env.get("working_dir", "")
            upload_path = existing_runtime_env.get("upload_path")
            if not is_dir_remote_uri(working_dir) and upload_path is None:
                workspace_id = os.environ["ANYSCALE_EXPERIMENTAL_WORKSPACE_ID"]
                cluster_id = os.environ["ANYSCALE_SESSION_ID"]
                new_runtime_env = upload_and_rewrite_workspace_working_dir(
                    api_client=self.api_client,
                    existing_runtime_env=existing_runtime_env,
                    workspace_id=workspace_id,
                    cluster_id=cluster_id,
                    log=self.log,
                )
                service_config.ray_serve_config["runtime_env"] = new_runtime_env

    def deploy_from_config(self, service_config: ServiceConfig):
        apply_service_config = format_service_config(service_config)
        # TODO: shawnp. Change to the new Services API
        # once the new endpoint works for both Service v1 and v2.
        create_production_service = apply_service_config.service_v1_config
        if create_production_service:
            service = self.api_client.apply_service_api_v2_decorated_ha_jobs_apply_service_put(
                create_production_service
            ).result
        else:
            # TODO: shawnp. This condition should never occur
            raise RuntimeError("The Service configuration is missing")
        service_id = service.id
        current_state = self._format_service_state(service.state.current_state)

        maximum_uptime_minutes = self._get_maximum_uptime_minutes(service)
        self.log.info(
            f"Maximum uptime is {self._get_maximum_uptime_output(maximum_uptime_minutes)} for clusters launched by this service."
            f"{self._get_additional_log_if_maximum_uptime_enabled(maximum_uptime_minutes)}"
        )

        self.log.info(
            f"Service {service_id} has been deployed. Current state of service: {current_state}."
        )
        self.log.info(
            f"Query the status of the service with `anyscale service list --service-id {service.id}`."
        )
        self.log.info(
            f'View the service in the UI at {get_endpoint(f"/services/{service.id}")}.'
        )

    def _get_maximum_uptime_minutes(self, service: ProductionJob) -> Optional[int]:
        compute_config: ComputeTemplate = self.api_client.get_compute_template_api_v2_compute_templates_template_id_get(
            service.config.compute_config_id
        ).result
        return compute_config.config.maximum_uptime_minutes

    def _get_maximum_uptime_output(self, maximum_uptime_minutes: Optional[int]) -> str:
        if maximum_uptime_minutes and maximum_uptime_minutes > 0:
            return f"set to {maximum_uptime_minutes} minutes"
        return "disabled"

    def _get_additional_log_if_maximum_uptime_enabled(
        self, maximum_uptime_minutes: Optional[int]
    ) -> str:
        if maximum_uptime_minutes and maximum_uptime_minutes > 0:
            return " This may cause disruptions. To disable, update the compute config."
        return ""

    def _populate_service_config(self, config_dict: Dict[str, Any]) -> ServiceConfig:
        """
        If the service is deployed from the workspace, we populate values from the workspace
        if the config doesn't specify them, such as cluster env and compute config.
        """
        if (
            "ANYSCALE_EXPERIMENTAL_WORKSPACE_ID" in os.environ
            and "ANYSCALE_SESSION_ID" in os.environ
        ):
            cluster = self.anyscale_api_client.get_cluster(
                os.environ["ANYSCALE_SESSION_ID"]
            ).result
            # If the job configs are not specified, infer them from the workspace:
            if "build_id" not in config_dict and "cluster_env" not in config_dict:
                config_dict["build_id"] = cluster.cluster_environment_build_id
            if "project_id" not in config_dict:
                config_dict["project_id"] = cluster.project_id
            if (
                "compute_config" not in config_dict
                and "compute_config_id" not in config_dict
            ):
                config_dict["compute_config_id"] = cluster.cluster_compute_id
        service_config = ServiceConfig.parse_obj(config_dict)
        return service_config

    def _read_service_config_file(self, service_config_file: str) -> ServiceConfig:
        if not os.path.exists(service_config_file):
            raise click.ClickException(f"Config file {service_config_file} not found.")

        with open(service_config_file) as f:
            config_dict = yaml.safe_load(f)

        return self._populate_service_config(config_dict)

    def rollout(
        self,
        service_config_file: str,
        name: Optional[str] = None,
        version: Optional[str] = None,
        canary_percent: Optional[int] = None,
        rollout_strategy: Optional[str] = None,
    ):
        """
        Deploys a Service 2.0.
        """
        config = self.generate_config_from_file(
            service_config_file,
            name=name,
            version=version,
            canary_percent=canary_percent,
            rollout_strategy=rollout_strategy,
        )

        service_v2_config = format_service_config_v2(config)
        service = self.api_client.apply_service_v2_api_v2_services_v2_apply_put(
            service_v2_config
        ).result

        current_state = self._format_service_state(service.goal_state)
        self.log.info(
            f"Service {service.id} has been deployed. Service is transitioning towards: {current_state}."
        )
        self.log.info(
            f'View the service in the UI at {get_endpoint(f"/services/{service.id}")}.'
        )

    def list(
        self,
        *,
        name: Optional[str] = None,
        service_id: Optional[str] = None,
        project_id: Optional[str] = None,
        created_by_me: bool = False,
        max_items: int = 10,
    ) -> None:
        try:
            self._list_via_service_list_endpoint(
                name=name,
                service_id=service_id,
                project_id=project_id,
                created_by_me=created_by_me,
                max_items=max_items,
            )
        except Exception as e:  # noqa: BLE001
            if "Services v2 is not enabled for your organization." in str(e):
                # An exception will be thrown if the "services_v2_enable_all_features" flag is not enabled
                # Therefore, we default to showing only Service v1.
                # TODO: shawnp. Remove the exception logic once we enable service v2 flag for all customers
                self.job_controller.list(
                    name=name,
                    job_id=service_id,
                    project_id=project_id,
                    include_all_users=not created_by_me,
                    include_archived=False,
                    max_items=max_items,
                    is_service=True,
                )
            else:
                self.log.info(
                    "We are unable to get a list of Services."
                    "This is unexpected and please contact Anyscale support for more info."
                )

    def _list_via_service_list_endpoint(
        self,
        *,
        name: Optional[str] = None,
        service_id: Optional[str] = None,
        project_id: Optional[str] = None,
        created_by_me: bool = False,
        max_items: int,
    ):
        services = []
        if service_id:
            if _is_v2_service_id(service_id):
                service_v2_result: DecoratedProductionServiceV2APIModel = (
                    self.api_client.get_service_api_v2_services_v2_service_id_get(
                        service_id
                    ).result
                )
                services.append(
                    [
                        service_v2_result.name,
                        service_v2_result.id,
                        self._format_service_state(service_v2_result.current_state),
                        # TODO: change to email once https://github.com/anyscale/product/pull/18189 is merged
                        service_v2_result.creator_id,
                    ]
                )
            else:
                service_v1_result: DecoratedProductionJob = self.api_client.get_job_api_v2_decorated_ha_jobs_production_job_id_get(
                    service_id
                ).result
                services.append(
                    [
                        service_v1_result.name,
                        service_v1_result.id,
                        self._format_service_state(
                            service_v1_result.state.current_state
                        ),
                        service_v1_result.creator.email,
                    ]
                )

        else:
            services_data = self._get_services_by_name(
                name=name,
                project_id=project_id,
                created_by_me=created_by_me,
                max_items=max_items,
            )

            services = [
                [
                    service.name,
                    service.id,
                    self._format_service_state(service.current_state),
                    service.creator.email,
                ]
                for service in services_data
            ]

        table = tabulate.tabulate(
            services,
            headers=["NAME", "ID", "CURRENT STATE", "CREATOR"],
            tablefmt="plain",
        )
        self.log.info(f'View your Services in the UI at {get_endpoint("/services")}')
        self.log.info(f"Services:\n{table}")

    def _format_service_state(
        self, state: Union[ServiceEventCurrentState, HaJobStates, ServiceGoalStates]
    ):
        if state == "RUNNING":
            return "ACTIVE"
        return state

    def archive(self, service_id: str):
        if _is_v2_service_id(service_id):
            raise click.ClickException(
                "archive is not currently supported for v2 services."
                "Please contact Anyscale support for more information."
            )

        self.job_controller.archive(service_id, None, is_service=True)

    def rollback(self, service_id: str) -> None:
        if not _is_v2_service_id(service_id):
            raise click.ClickException(
                "rollback is only supported for v2 services."
                "Please contact Anyscale support for more information."
            )

        service = self.api_client.rollback_service_api_v2_services_v2_service_id_rollback_post(
            service_id
        ).result

        self.log.info(f"Service {service.id} rollback initiated.")
        self.log.info(
            f'View the service in the UI at {get_endpoint(f"/services/{service.id}")}.'
        )

    def terminate(self, service_id: str) -> None:
        if _is_v2_service_id(service_id):
            self.api_client.terminate_service_api_v2_services_v2_service_id_terminate_post(
                service_id
            )
        else:
            self.job_controller.terminate(service_id, None, is_service=True)



from click import ClickException
from google.api_core.exceptions import NotFound
import pytest

from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.models.gcp_file_store_config import (
    GCPFileStoreConfig,
)
from anyscale.utils.gcp_utils import get_gcp_filestore_config


@pytest.mark.parametrize(
    ("filestore_id", "response", "not_found", "expected_result"),
    [
        pytest.param(
            "regional-filestore",
            ("200 OK", "regional_filestore.json"),
            False,
            GCPFileStoreConfig(
                instance_name="projects/anyscale-bridge-deadbeef/locations/us-central1/instances/regional-filestore",
                root_dir="something",
                mount_target_ip="172.22.236.2",
            ),
            id="regional",
        ),
        pytest.param(
            "zonal-filestore",
            ("200 OK", "zonal_filestore.json"),
            False,
            GCPFileStoreConfig(
                instance_name="projects/anyscale-bridge-deadbeef/locations/us-central1/instances/zonal-filestore",
                root_dir="test",
                mount_target_ip="172.27.155.250",
            ),
            id="zonal",
        ),
        pytest.param(
            "regional-filestore",
            ("200 OK", "regional_filestore_wrong_vpc.json"),
            False,
            None,
            id="wrong_vpc",
        ),
        pytest.param(
            "regional-filestore", ("404 Not Found", None), True, None, id="not_found"
        ),
    ],
)
def test_get_gcp_filestore_config(
    setup_mock_server, filestore_id, capsys, response, not_found, expected_result,
):
    factory, tracker = setup_mock_server
    mock_project_id = "anyscale-bridge-deadbeef"
    mock_vpc_name = "vpc_one"
    mock_filestore_region = "us-central1"
    tracker.reset({".*": [response]})
    if expected_result:
        assert (
            get_gcp_filestore_config(
                factory,
                mock_project_id,
                mock_vpc_name,
                mock_filestore_region,
                filestore_id,
                BlockLogger(),
            )
            == expected_result
        )
    else:
        if not_found:
            with pytest.raises(NotFound):
                get_gcp_filestore_config(
                    factory,
                    mock_project_id,
                    mock_vpc_name,
                    mock_filestore_region,
                    filestore_id,
                    BlockLogger(),
                )
            _, err = capsys.readouterr()
            assert "Could not find Filestore with id" in err
        else:
            with pytest.raises(ClickException):
                get_gcp_filestore_config(
                    factory,
                    mock_project_id,
                    mock_vpc_name,
                    mock_filestore_region,
                    filestore_id,
                    BlockLogger(),
                )
            _, err = capsys.readouterr()
            assert "This cannot be edited on an existing Filestore instance." in err



import ipaddress
import json
from pathlib import Path
import re
import sys
from tempfile import NamedTemporaryFile
from typing import Any, Callable, List

from google.api_core.exceptions import NotFound
from google.cloud.compute_v1.types import Subnetwork
from google.cloud.compute_v1.types.compute import FirewallPolicyRule
from google.iam.v1.policy_pb2 import Binding
from googleapiclient.errors import HttpError
import pytest

from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.models import CreateCloudResourceGCP
from anyscale.client.openapi_client.models.gcp_file_store_config import (
    GCPFileStoreConfig,
)
from anyscale.gcp_verification import (
    _check_bucket_region,
    _firewall_rules_from_proto_resp,
    _gcp_subnet_has_enough_capacity,
    _verify_service_account_on_bucket,
    GCPLogger,
    verify_cloud_storage,
    verify_filestore,
    verify_firewall_policy,
    verify_gcp_access_service_account,
    verify_gcp_dataplane_service_account,
    verify_gcp_networking,
    verify_gcp_project,
)
from anyscale.utils.network_verification import Direction, FirewallRule, Protocol


_MOCK_GCP_PROJECT_ID = "anyscale-bridge-deadbeef"

_MOCK_GCP_CLOUD_REGION = "us-west1"


def generate_gcp_cloud_mock_resources(**overrides: Any) -> CreateCloudResourceGCP:
    default_resources = {
        "gcp_vpc_id": "vpc_one",
        "gcp_anyscale_iam_service_account_email": "anyscale-admin@anyscale-bridge-deadbeef.iam.gserviceaccount.com",
        "gcp_cluster_node_service_account_email": "cld-xyz-access@anyscale-bridge-deadbeef.iam.gserviceaccount.com",
        "gcp_subnet_ids": ["subnet_one"],
        "gcp_firewall_policy_ids": ["firewall"],
        "gcp_filestore_config": GCPFileStoreConfig(
            instance_name="projects/anyscale-bridge-deadbeef/locations/us-west1/instances/store-one",
            root_dir="mock_root_dir",
            mount_target_ip="mock_ip",
        ),
        "gcp_cloud_storage_bucket_id": "cloud-bucket",
    }
    return CreateCloudResourceGCP(**{**default_resources, **overrides})


@pytest.mark.parametrize(
    ("responses", "expected_result", "num_calls"),
    [
        pytest.param(
            {".*/networks/.*": [("404 Not Found", None)]}, False, 1, id="Network404",
        ),
        pytest.param(
            {
                ".*/networks/.*": [("200 OK", "networks.json")],
                ".*/subnetworks/.*": [("404 Not Found", None)],
            },
            False,
            2,
            id="SubNet404",
        ),
        pytest.param(
            {
                ".*/networks/.*": [("200 OK", "networks.json")],
                ".*/subnetworks/.*": [("200 OK", "subnetworks.json")],
            },
            True,
            2,
            id="Success",
        ),
        pytest.param(
            {
                ".*/networks/.*": [("200 OK", "networks.json")],
                ".*/subnetworks/.*": [("200 OK", "subnetworks_other.json")],
            },
            False,
            2,
            id="WrongNetwrork",
        ),
    ],
)
def test_verify_gcp_networking(
    setup_mock_server, responses, expected_result: bool, num_calls: int
):
    factory, tracker = setup_mock_server
    tracker.reset(responses=responses)
    assert (
        verify_gcp_networking(
            factory,
            generate_gcp_cloud_mock_resources(),
            _MOCK_GCP_PROJECT_ID,
            _MOCK_GCP_CLOUD_REGION,
            _gcp_logger(),
        )
        == expected_result
    )
    assert len(tracker.seen_requests) == num_calls, tracker.seen_requests


@pytest.mark.parametrize(
    ("cidr", "result", "log"),
    [
        ("10.0.0.0/16", True, ""),
        ("10.0.0.0/19", True, ""),
        ("10.0.0.0/21", True, "We suggest at least"),
        ("10.0.0.0/25", False, "We want at least"),
        pytest.param(
            "",
            False,
            "",
            marks=pytest.mark.xfail(raises=ValueError, strict=True),
            id="BadIPAddress",
        ),
    ],
)
def test_gcp_subnet_has_enough_capacity(cidr: str, result: bool, log: str, capsys):
    subnet = Subnetwork(ip_cidr_range=cidr)

    assert _gcp_subnet_has_enough_capacity(subnet, BlockLogger()) == result

    stdout, stderr = capsys.readouterr()
    sys.stdout.write(stdout)
    sys.stderr.write(stderr)

    if log:
        assert re.search(log, stderr)


@pytest.mark.parametrize(
    "call_factory",
    [
        pytest.param(
            lambda f: f.compute_v1.NetworksClient().get(project="abc", network="cde"),
            id="Compute",
        ),
        pytest.param(
            lambda f: f.resourcemanager_v3.ProjectsClient().get_project(
                name="projects/abc"
            ),
            id="ResourceManager",
        ),
    ],
)
def test_client_factory_cloud_client(setup_mock_server, call_factory: Callable):
    factory, tracker = setup_mock_server
    tracker.reset(responses={".*": [("404 Not Found", None)]})
    with pytest.raises(NotFound):
        call_factory(factory)


def test_client_factory_apis(setup_mock_server):
    factory, tracker = setup_mock_server
    tracker.reset(responses={".*": [("418 I'm a teapot", None)]})

    with pytest.raises(HttpError) as e:
        factory.build("iam", "v1").projects().serviceAccounts().get(
            name="projects/-/serviceAccounts/abc"
        ).execute()
    assert e.value.status_code == 418


@pytest.mark.parametrize(
    ("responses", "expected_result"),
    [
        pytest.param({".*": [("404 Not Found", None)]}, False, id="NotFound"),
        pytest.param(
            {
                ".*": [
                    ("200 OK", "project_get.json"),
                    ("200 OK", "project_iam_bindings.json"),
                ]
            },
            True,
            id="ProjectExists",
        ),
        pytest.param(
            {".*": [("200 OK", "project_get_inactive.json")]},
            False,
            id="ProjectInactive",
        ),
    ],
)
def test_verify_gcp_project(setup_mock_server, responses, expected_result: bool):
    factory, tracker = setup_mock_server
    tracker.reset(responses)
    assert (
        verify_gcp_project(factory, _MOCK_GCP_PROJECT_ID, _gcp_logger())
        == expected_result
    )


def test_verify_gcp_missing_bindigs(setup_mock_server, capsys):
    factory, tracker = setup_mock_server
    tracker.reset(
        {
            ".*": [
                ("200 OK", "project_get.json"),
                ("200 OK", "project_iam_bindings.json"),
            ]
        }
    )
    assert verify_gcp_project(factory, _MOCK_GCP_PROJECT_ID, _gcp_logger())
    _, err = capsys.readouterr()
    assert (
        "The Compute Engine Service Agent does not have the standard IAM Role of 'roles/compute.serviceAgent'"
        in err
    )


@pytest.mark.parametrize(
    ("service_account", "expected_result"),
    [
        pytest.param(
            "anyscale-admin-owner@anyscale-bridge-deadbeef.iam.gserviceaccount.com",
            True,
            id="Owner",
        ),
        pytest.param(
            "anyscale-admin-editor@anyscale-bridge-deadbeef.iam.gserviceaccount.com",
            True,
            id="Editor",
        ),
        pytest.param(
            "anyscale-admin-editor-not-signer@anyscale-bridge-deadbeef.iam.gserviceaccount.com",
            False,
            id="EditorNotSigner",
        ),
        pytest.param(
            "nope@anyscale-bridge-deadbeef.iam.gserviceaccount.com",
            False,
            id="NotGranted",
        ),
    ],
)
def test_verify_gcp_access_service_account(
    setup_mock_server, service_account: str, expected_result: bool
):
    factory, tracker = setup_mock_server
    tracker.reset(
        {
            ".*": [
                ("200 OK", "access_service_account_permissions.json"),
                ("200 OK", "project_iam_binding_access.json"),
            ]
        }
    )
    mock_resources = generate_gcp_cloud_mock_resources(
        gcp_anyscale_iam_service_account_email=service_account
    )
    assert (
        verify_gcp_access_service_account(
            factory, mock_resources, _MOCK_GCP_PROJECT_ID, _gcp_logger(),
        )
        == expected_result
    )


@pytest.mark.parametrize(
    ("responses", "expected_result", "output"),
    [
        pytest.param(
            {".*": [("404 Not Found", None), ("404 Not Found", None)]},
            False,
            re.compile(".*"),
            id="Neither Exist",
        ),
        pytest.param(
            {
                ".*": [
                    ("200 OK", "global_firewall.json"),
                    ("200 OK", "subnetworks.json"),
                ],
            },
            True,
            re.compile(".*"),
            id="GlobalFirewall",
        ),
        pytest.param(
            {
                ".*": [
                    ("404 Not Found", None),
                    ("200 OK", "regional_firewall.json"),
                    ("200 OK", "subnetworks.json"),
                ],
            },
            True,
            re.compile("Global firweall policy .* not found"),
            id="RegionalFirewall",
        ),
        pytest.param(
            {".*": [("200 OK", "global_firewall_wrong_vpc.json")],},
            False,
            re.compile(".*is not associated with the VPC.*"),
            id="WrongNetwork",
        ),
    ],
)
def test_verify_firewall_policy(
    setup_mock_server, responses, expected_result: bool, output: re.Pattern, capsys
):
    factory, tracker = setup_mock_server
    tracker.reset(responses=responses)

    assert (
        verify_firewall_policy(
            factory,
            generate_gcp_cloud_mock_resources(),
            _MOCK_GCP_PROJECT_ID,
            _MOCK_GCP_CLOUD_REGION,
            GCPLogger(BlockLogger(), "project"),
        )
        == expected_result
    )
    _, stderr = capsys.readouterr()
    assert output.search(stderr), f"Missing output in {stderr}"


@pytest.mark.parametrize(
    ("proto", "rule"),
    [
        (FirewallPolicyRule(action="deny"), []),
        (
            FirewallPolicyRule(
                direction="EGRESS",
                action="allow",
                match={"src_ip_ranges": ["10.10.10.0/24"]},
            ),
            [
                FirewallRule(
                    direction=Direction.EGRESS,
                    protocol=Protocol.all,
                    network=ipaddress.ip_network("10.10.10.0/24"),
                    ports=None,
                )
            ],
        ),
        (
            FirewallPolicyRule(
                direction="EGRESS",
                action="allow",
                match={
                    "src_ip_ranges": ["10.10.10.0/24"],
                    "layer4_configs": [{"ip_protocol": "tcp"}],
                },
            ),
            [
                FirewallRule(
                    direction=Direction.EGRESS,
                    protocol=Protocol.tcp,
                    network=ipaddress.ip_network("10.10.10.0/24"),
                    ports=None,
                )
            ],
        ),
    ],
)
def test_firewall_rules_from_proto_resp(
    proto: FirewallPolicyRule, rule: List[FirewallRule]
):
    assert _firewall_rules_from_proto_resp([proto]) == rule
    assert _firewall_rules_from_proto_resp([proto, proto]) == rule * 2


def test_verify_gcp_access_service_account_does_not_exist(setup_mock_server):
    factory, tracker = setup_mock_server
    tracker.reset({".*": [("404 Not Found", None)]})
    assert not verify_gcp_access_service_account(
        factory,
        generate_gcp_cloud_mock_resources(),
        _MOCK_GCP_PROJECT_ID,
        _gcp_logger(),
    )


@pytest.mark.parametrize(
    ("responses", "expected_result"),
    [
        pytest.param({".*": [("404 Not Found", None)]}, False, id="DoesNotExist"),
        pytest.param(
            {
                ".*": [
                    ("200 OK", "dataplane_service_account.json"),
                    ("200 OK", "project_iam_binding_access.json"),
                ]
            },
            True,
            id="DoesNotExist",
        ),
    ],
)
def test_verify_gcp_dataplane_service_account(
    setup_mock_server, responses, expected_result: bool, capsys
):
    factory, tracker = setup_mock_server
    tracker.reset(responses)
    assert (
        verify_gcp_dataplane_service_account(
            factory,
            generate_gcp_cloud_mock_resources(),
            _MOCK_GCP_PROJECT_ID,
            _gcp_logger(),
        )
        == expected_result
    )

    _, err = capsys.readouterr()
    if expected_result:
        # Check if there is a warning about missing artifact Registry warnings if we pass verification.
        assert re.search("roles/artifactregistry.reader`", err) and re.search(
            "safe to ignore if you are not using", err
        )


def test_verify_gcp_dataplane_service_account_wrong_project(setup_mock_server, capsys):
    factory, tracker = setup_mock_server
    tracker.reset(
        {
            ".*": [
                ("200 OK", "dataplane_service_account.json"),
                ("200 OK", "project_iam_binding_access.json"),
            ]
        }
    )
    assert verify_gcp_dataplane_service_account(
        factory, generate_gcp_cloud_mock_resources(), "something-else", _gcp_logger(),
    )

    _, stderr = capsys.readouterr()
    assert (
        "constraints/iam.disableCrossProjectServiceAccountUsage` is not enforced"
        in stderr
    )


@pytest.mark.parametrize(
    ("response", "result"),
    [
        pytest.param(("200 OK", "regional_filestore.json"), True, id="regional"),
        pytest.param(("200 OK", "zonal_filestore.json"), True, id="zonal"),
        pytest.param(
            ("200 OK", "regional_filestore_wrong_vpc.json"), False, id="wrong_vpc"
        ),
        pytest.param(("404 Not Found", None), False, id="not_found"),
    ],
)
def test_verify_filestore(setup_mock_server, response, result: bool):
    factory, tracker = setup_mock_server
    tracker.reset({".*": [response]})
    assert (
        verify_filestore(factory, generate_gcp_cloud_mock_resources(), _gcp_logger(),)
        == result
    )


@pytest.mark.parametrize(
    ("response", "has_warning"),
    [
        pytest.param(("200 OK", "regional_filestore.json"), False, id="Enterprise"),
        pytest.param(("200 OK", "zonal_filestore.json"), True, id="NonEnterprise"),
    ],
)
def test_verify_filestore_warn(setup_mock_server, capsys, response, has_warning: bool):
    factory, tracker = setup_mock_server
    tracker.reset({".*": [response]})
    assert verify_filestore(factory, generate_gcp_cloud_mock_resources(), _gcp_logger())
    _, err = capsys.readouterr()
    assert ("ENTERPRISE" in err) == has_warning


@pytest.mark.parametrize(
    ("responses", "expected_result"),
    [
        pytest.param([("404 Not Found", None)], False, id="NoBucket"),
        pytest.param(
            [
                ("200 OK", "storage_bucket.json"),
                ("200 OK", "storage_bucket_policy.json"),
            ],
            True,
            id="Success",
        ),
    ],
)
def test_verify_cloud_storage(setup_mock_server, responses, expected_result: bool):
    factory, tracker = setup_mock_server
    tracker.reset({".*": responses})
    assert (
        verify_cloud_storage(
            factory,
            generate_gcp_cloud_mock_resources(),
            _MOCK_GCP_PROJECT_ID,
            _MOCK_GCP_CLOUD_REGION,
            _gcp_logger(),
        )
        == expected_result
    )


@pytest.mark.parametrize(
    ("update_fn", "expected_warn"),
    [
        pytest.param(
            lambda x: x["iamConfiguration"]["uniformBucketLevelAccess"].update(
                {"enabled": False}
            ),
            "does not have Uniform Bucket Access enabled,",
            id="UniformBucketAccess",
        ),
        pytest.param(
            lambda x: x["iamConfiguration"].update(
                {"publicAccessPrevention": "unspecified"}
            ),
            "public access prevention",
            id="PublicAccessPrevention",
        ),
        pytest.param(
            lambda x: x.update({"cors": []}), "create the correct CORS rule", id="CORS"
        ),
    ],
)
def test_verify_cloud_storage_configuration_warnings(
    setup_mock_server, update_fn: Callable, expected_warn: str, capsys
):
    factory, tracker = setup_mock_server

    with NamedTemporaryFile(mode="w") as bad_bucket:
        responses = [
            ("200 OK", "storage_bucket.json"),
            ("200 OK", "storage_bucket_policy.json"),
            ("200 OK", bad_bucket.name),
            ("200 OK", "storage_bucket_policy.json"),
        ]
        with open(
            Path(__file__).parent.joinpath("gcp_responses", "storage_bucket.json")
        ) as f:
            existing = json.load(f)
            update_fn(existing)
            bad_bucket.write(json.dumps(existing))
            bad_bucket.flush()

        tracker.reset({".*": responses})
        assert verify_cloud_storage(
            factory,
            generate_gcp_cloud_mock_resources(),
            _MOCK_GCP_PROJECT_ID,
            _MOCK_GCP_CLOUD_REGION,
            _gcp_logger(),
        )

        # First call should not have an error
        _, err = capsys.readouterr()
        assert expected_warn not in err

        # Second call should have the error
        assert verify_cloud_storage(
            factory,
            generate_gcp_cloud_mock_resources(),
            _MOCK_GCP_PROJECT_ID,
            _MOCK_GCP_CLOUD_REGION,
            _gcp_logger(),
        )
        _, err = capsys.readouterr()
        assert expected_warn in err


def test_verify_cloud_storage_improper_service_account_configuration(
    setup_mock_server, capsys
):
    factory, tracker = setup_mock_server
    tracker.reset(
        {
            ".*": [
                ("200 OK", "storage_bucket.json"),
                ("200 OK", "storage_bucket_policy.json"),
            ]
            * 2
        }
    )

    common_error_text = (
        "other_email@project.com requires the following permissions on Bucket "
    )
    mock_resource = generate_gcp_cloud_mock_resources(
        gcp_anyscale_iam_service_account_email="other_email@project.com"
    )
    assert verify_cloud_storage(
        factory,
        mock_resource,
        _MOCK_GCP_PROJECT_ID,
        _MOCK_GCP_CLOUD_REGION,
        _gcp_logger(),
    )
    _, err = capsys.readouterr()
    assert err.count(common_error_text) == 1

    mock_resource = generate_gcp_cloud_mock_resources(
        gcp_cluster_node_service_account_email="other_email@project.com"
    )
    assert verify_cloud_storage(
        factory,
        mock_resource,
        _MOCK_GCP_PROJECT_ID,
        _MOCK_GCP_CLOUD_REGION,
        _gcp_logger(),
    )

    _, err = capsys.readouterr()
    assert err.count(common_error_text) == 1


@pytest.mark.parametrize(
    ("bindings", "expected_result"),
    [
        pytest.param(
            [Binding(role="roles/storage.admin", members=["sa@email.com"])],
            True,
            id="Admin",
        ),
        pytest.param(
            [
                Binding(role="roles/storage.objectAdmin", members=["sa@email.com"]),
                Binding(
                    role="roles/storage.legacyBucketReader", members=["sa@email.com"]
                ),
            ],
            True,
            id="Legacy+Admin",
        ),
        pytest.param(
            [Binding(role="roles/storage.legacyBucketOwner", members=["sa@email.com"])],
            False,
            id="MissingObjectRead",
        ),
        pytest.param(
            [
                Binding(
                    role="roles/storage.legacyBucketOwner", members=["sa@email.com"]
                ),
                Binding(role="roles/storage.objectAdmin", members=["sa_two@email.com"]),
            ],
            False,
            id="NecessaryPermissionOnCorrect",
        ),
    ],
)
def test_verify_service_account_on_bucket(bindings: List[Binding], expected_result):
    assert (
        _verify_service_account_on_bucket("sa@email.com", bindings) == expected_result
    )
    assert not _verify_service_account_on_bucket("dne@email.com", bindings)


@pytest.mark.parametrize(
    ("region", "file", "result"),
    [
        ("us-west1", "storage_bucket.json", True),
        ("us-east1", "storage_bucket.json", True),
        ("europe-central2", "storage_bucket.json", False),
        ("europe-central2", "storage_dual_region.json", True),
        ("europe-central1", "storage_dual_region.json", False),
        ("southamerica-west1", "storage_single_region.json", True),
        ("southamerica-east1", "storage_single_region.json", False),
    ],
)
def test_check_bucket_region(setup_mock_server, region: str, file: str, result: bool):
    factory, tracker = setup_mock_server
    tracker.reset({".*": [("200 OK", file)]})
    assert (
        _check_bucket_region(
            factory.storage.Client("project").get_bucket("bucket"), region
        )[0]
        == result
    )


def _gcp_logger() -> GCPLogger:
    return GCPLogger(BlockLogger(), "project_abc", yes=True)



import asyncio
from http import HTTPStatus
from typing import List, Optional

from aiohttp import ClientConnectionError, ClientResponseError

from anyscale.api_utils.exceptions.log_retrieval_errors import (
    LogNotFoundOnActiveClusterError,
    UnsupportedLogRetrievalMethodError,
    wrap_as_unsupported_log_retrieval_method_error,
)
from anyscale.api_utils.job_util import _get_job_run_id
from anyscale.api_utils.logs_util import (
    _download_log_from_ray_json_response,
    _download_logs_concurrently,
    _remove_ansi_escape_sequences,
)
from anyscale.controllers.logs_controller import DEFAULT_PARALLELISM
from anyscale.sdk.anyscale_client.api.default_api import DefaultApi as BaseApi
from anyscale.sdk.anyscale_client.models.log_download_result import LogDownloadResult


async def _get_logs_from_running_job(
    base_api: BaseApi, job_run_id: str, remove_escape_chars: bool = True
) -> str:
    with wrap_as_unsupported_log_retrieval_method_error():
        job_logs_url: str = base_api.get_job_logs_stream(
            job_id=job_run_id
        ).result.http_url
    try:
        logs = await _download_log_from_ray_json_response(job_logs_url)
        if remove_escape_chars:
            logs = _remove_ansi_escape_sequences(logs)
        return logs
    except (ClientConnectionError, asyncio.TimeoutError):
        raise UnsupportedLogRetrievalMethodError(
            f'Streaming logs is only supported if the Anyscale SDK client can reach the cluster running the job "{job_run_id}". '
            "Check that the client is running on the same private network (may be protected behind VPN)."
        )
    except ClientResponseError as e:
        # Will return 404 if job run (Ray job) is not found
        # https://github.com/ray-project/ray/blob/6334576223dd916469fcdac185752b80e0c3e416/dashboard/modules/job/job_head.py#L485
        if e.status == HTTPStatus.NOT_FOUND:
            raise LogNotFoundOnActiveClusterError(
                f'Job "{job_run_id}" is not a part of the current active session of the cluster. '
                "The cluster may have restarted since the job ran."
            )
        raise e


async def _get_job_logs_from_storage_bucket(
    base_api: BaseApi,
    *,
    job_id: Optional[str] = None,
    job_run_id: Optional[str] = None,
    parallelism: int = DEFAULT_PARALLELISM,
    remove_escape_chars: bool = True,
) -> str:
    with wrap_as_unsupported_log_retrieval_method_error():
        job_run_id = _get_job_run_id(base_api, job_id=job_id, job_run_id=job_run_id)
        all_log_chunk_urls: List[str] = []

        log_download_result: LogDownloadResult = base_api.get_job_logs_download(
            job_id=job_run_id, all_logs=True
        ).result
        for log_chunk in log_download_result.log_chunks:
            all_log_chunk_urls.append(log_chunk.chunk_url)
    logs = await _download_logs_concurrently(all_log_chunk_urls, parallelism)
    if remove_escape_chars:
        logs = _remove_ansi_escape_sequences(logs)
    return logs


async def _get_logs_from_running_production_job(
    base_api: BaseApi, production_job_id: str, remove_escape_chars: bool = True,
) -> str:
    with wrap_as_unsupported_log_retrieval_method_error():
        last_job_run_id = _get_job_run_id(base_api, job_id=production_job_id)
    return await _get_logs_from_running_job(
        base_api, last_job_run_id, remove_escape_chars
    )



import subprocess
from typing import Dict, List, Optional, Set, Tuple, Union

from click import ClickException
from google.api_core.exceptions import NotFound
import google.auth
from google.auth.credentials import Credentials
from google.iam.v1.policy_pb2 import Binding

from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.models.gcp_file_store_config import (
    GCPFileStoreConfig,
)


def get_application_default_credentials(
    logger: BlockLogger,
) -> Tuple[Credentials, Optional[str]]:
    """Get application default credentials, or run `gcloud` to try to log in."""
    try:
        return google.auth.default()
    except google.auth.exceptions.DefaultCredentialsError as e:
        logger.warning(
            "Could not automatically determine Google Application Default Credentials, trying to authenticate via GCloud"
        )
        auth_login = subprocess.run(["gcloud", "auth", "application-default", "login"])
        if auth_login.returncode != 0:
            raise RuntimeError("Failed to authenticate via gcloud") from e

        return google.auth.default()


def binding_from_dictionary(
    inp: List[Dict[str, Union[List[str], str]]]
) -> List[Binding]:
    return [Binding(role=b["role"], members=b["members"]) for b in inp]


def check_policy_bindings(
    iam_policy: List[Binding], member: str, possible_roles: Set[str]
) -> bool:
    """Checks if `member` has any role in `possible_roles` given the specified iam_policy."""
    return any(
        policy.role in possible_roles and member in policy.members
        for policy in iam_policy
    )


def check_required_policy_bindings(
    iam_policy: List[Binding], member: str, required_roles: Set[str]
) -> bool:
    """Checks if `member` has all roles in `required_roles` given the specified iam_policy."""
    granted_roles = {policy.role for policy in iam_policy if member in policy.members}
    return required_roles.issubset(granted_roles)


def get_gcp_filestore_config(
    factory, project_id, vpc_name, filestore_location, filestore_instance_id, logger
):
    client = factory.filestore_v1.CloudFilestoreManagerClient()
    instance_name = "projects/{}/locations/{}/instances/{}".format(
        project_id, filestore_location, filestore_instance_id
    )
    try:
        file_store = client.get_instance(name=instance_name)
    except NotFound as e:
        logger.error(
            f"Could not find Filestore with id {filestore_instance_id} in project {project_id}. Please validate that you're using the correct GCP project and that the resource values are correct."
        )
        raise e
    root_dir = file_store.file_shares[0].name
    for v in file_store.networks:
        if v.network == vpc_name:
            mount_target_ip = v.ip_addresses[0]
            break
    else:
        logger.error(
            f"Filestore {filestore_instance_id} is not connected to {vpc_name}, but to {file_store.networks}. "
            f"This cannot be edited on an existing Filestore instance. Please recreate the filestore and connect it to {vpc_name}."
        )
        raise ClickException(
            f"Filestore {filestore_instance_id} is not connected to {vpc_name}."
        )
    return GCPFileStoreConfig(
        instance_name=instance_name, root_dir=root_dir, mount_target_ip=mount_target_ip,
    )



from typing import List, Optional, Tuple

import click

from anyscale.controllers.exec_controller import ExecController


@click.command(
    name="exec",
    hidden=True,
    help="[DEPRECATED] Execute shell commands in interactive cluster.",
)
@click.option(
    "--session-name",
    "-n",
    type=str,
    required=False,
    default=None,
    help="Cluster name: optional if only one running cluster.",
    envvar="SESSION_NAME",
    hidden=True,
)
@click.option(
    "--cluster-name",
    "-n",
    type=str,
    required=False,
    default=None,
    help="Cluster name: optional if only one running cluster.",
    envvar="CLUSTER_NAME",
)
@click.option(
    "--screen", is_flag=True, default=False, help="Run the command in a screen."
)
@click.option("--tmux", is_flag=True, default=False, help="Run the command in tmux.")
@click.option(
    "--port-forward",
    "-p",
    required=False,
    multiple=True,
    type=int,
    help="Port to forward. Use this multiple times to forward multiple ports.",
)
@click.option(
    "--sync",
    is_flag=True,
    default=False,
    help="Rsync all the file mounts before executing the command.",
)
@click.option(
    "--stop",
    help="Stop session after command finishes executing.",
    is_flag=True,
    default=False,
)
@click.option(
    "--terminate",
    help="Terminate session after command finishes executing.",
    is_flag=True,
    default=False,
)
@click.argument("commands", nargs=-1, type=str)
def anyscale_exec(  # noqa: PLR0913
    session_name: str,
    cluster_name: Optional[str],
    screen: bool,
    tmux: bool,
    port_forward: Tuple[int],
    sync: bool,
    stop: bool,
    terminate: bool,
    commands: List[str],
) -> None:
    exec_controller = ExecController()
    exec_controller.anyscale_exec(
        cluster_name or session_name,
        screen,
        tmux,
        port_forward,
        sync,
        stop,
        terminate,
        commands,
    )



# pylint:disable=private-import
import copy
import hashlib
import logging
import os
from pathlib import Path
import tempfile
from typing import Any, Callable, Dict, Optional, TYPE_CHECKING
from urllib.parse import urlparse
import uuid

import click

from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.api.default_api import DefaultApi
from anyscale.client.openapi_client.models.cloud_providers import CloudProviders
from anyscale.client.openapi_client.models.cloud_with_cloud_resource import (
    CloudWithCloudResource,
)
from anyscale.client.openapi_client.models.decorated_session import DecoratedSession
from anyscale.client.openapi_client.models.user_info import UserInfo
from anyscale.shared_anyscale_utils.aws import AwsArn, S3_BUCKET_ARN_PREFIX
from anyscale.util import is_anyscale_workspace
from anyscale.utils.ray_utils import zip_directory  # type: ignore


if TYPE_CHECKING:
    pass

logger = logging.getLogger(__name__)


def _upload_file_to_google_cloud_storage(file: str, bucket: str, object_name: str):
    try:
        from google.cloud import storage

    except Exception:  # noqa: BLE001
        raise click.ClickException(
            "Could not upload file to Google Storage. Could not import the Google Storage Python API via `from google.cloud import storage`.  Please check your installation or try running `pip install --upgrade google-cloud-storage`."
        )
    try:
        storage_client = storage.Client()
        bucket_obj = storage_client.bucket(bucket)
        blob = bucket_obj.blob(object_name)
        blob.upload_from_filename(file)
    except Exception as e:  # noqa: BLE001
        raise click.ClickException(
            f"Could not upload file to Google Cloud Storage. Please check your credentials and ensure that the bucket exists. {e}"
        )


def _upload_file_to_s3(file: str, bucket: str, object_key: str):
    try:
        import boto3
    except Exception:  # noqa: BLE001
        raise click.ClickException(
            "Could not upload file to S3: Could not import the Amazon S3 Python API via `import boto3`.  Please check your installation or try running `pip install boto3`."
        )
    try:
        s3_client = boto3.client("s3")
        s3_client.upload_file(file, bucket, object_key)
    except Exception as e:  # noqa: BLE001
        raise click.ClickException(
            f"Could not upload file to S3. Check your credentials and that the bucket exists. Original error: {e}"
        )


def _get_remote_storage_object_name(upload_path, upload_filename):
    # Strip leading slash, otherwise bucket will create a new directory called "/".
    object_name = os.path.join(urlparse(upload_path).path, upload_filename).lstrip("/")
    return object_name


def _upload_file_to_remote_storage(
    source_file: str, upload_path: str, upload_filename: str
):
    parsed_upload_path = urlparse(upload_path)
    service = parsed_upload_path.scheme
    bucket = parsed_upload_path.netloc
    object_name = _get_remote_storage_object_name(upload_path, upload_filename)
    if service == "s3":
        _upload_file_to_s3(source_file, bucket, object_key=object_name)
    if service == "gs":
        _upload_file_to_google_cloud_storage(
            source_file, bucket, object_name=object_name
        )

    final_uploaded_filepath = os.path.join(upload_path, upload_filename)
    try:
        from smart_open import open

        open(final_uploaded_filepath)
    except Exception as e:  # noqa: BLE001
        raise click.ClickException(
            f"Could not open uploaded file, maybe something went wrong while uploading: {e}."
        )

    return final_uploaded_filepath


def is_dir_remote_uri(working_dir: str) -> bool:
    parsed = urlparse(working_dir)
    if parsed.scheme:
        return True
    return False


def upload_and_rewrite_working_dir(
    runtime_env_json: Dict[str, Any],
    upload_file_to_remote_storage_fn: Callable[
        [str, str, str], str
    ] = _upload_file_to_remote_storage,
) -> Dict[str, Any]:
    """Upload a local working_dir and rewrite the working_dir field with the destination remote URI.

    After uploading, deletes the "upload_path" field because it is no longer used and is not a valid
    OSS runtime env field.
    """
    if runtime_env_json.get("working_dir", None) is None:
        return runtime_env_json

    working_dir = runtime_env_json["working_dir"]
    if is_dir_remote_uri(working_dir):
        # The working dir is a remote URI already
        return runtime_env_json

    upload_path = runtime_env_json["upload_path"]
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_zip_file_path = os.path.join(
            temp_dir, "anyscale_generated_working_dir.zip"
        )
        zip_directory(
            working_dir,
            excludes=runtime_env_json.get("excludes", []),
            output_path=temp_zip_file_path,
            # Ray requires remote Zip URIs to consist of a single top-level directory when unzipped.
            include_parent_dir=True,
        )

        hash_val = hashlib.md5(Path(temp_zip_file_path).read_bytes()).hexdigest()
        uploaded_zip_file_name = f"_anyscale_pkg_{hash_val}.zip"
        final_uploaded_filepath = upload_file_to_remote_storage_fn(
            temp_zip_file_path, upload_path, uploaded_zip_file_name,
        )

    final_runtime_env = runtime_env_json.copy()
    final_runtime_env["working_dir"] = final_uploaded_filepath
    del final_runtime_env["upload_path"]
    return final_runtime_env


def override_runtime_env_for_local_working_dir(
    runtime_env: Optional[Dict[str, Any]]
) -> Optional[Dict[str, Any]]:
    """
    If runtime_env contains a "upload_path" field, we need to upload it first
    before backing it up inside EFS (if in workspaces).
    """
    if runtime_env:
        runtime_env = upload_and_rewrite_working_dir(runtime_env)

    if is_anyscale_workspace():
        import anyscale

        runtime_env = anyscale.snapshot_util.env_hook(runtime_env)
        # Snapshot the working directory to EFS so the job we submit can read it from there
        parsed_working_dir = urlparse(runtime_env["working_dir"])
        # We add prefix "file://" only if the working dir is local dir
        if not parsed_working_dir.scheme:
            runtime_env["working_dir"] = "file://" + runtime_env["working_dir"]
        fake_job_id = uuid.uuid4().hex
        runtime_env = anyscale.snapshot_util.checkpoint_job(fake_job_id, runtime_env)
        runtime_env["working_dir"] = "file://" + runtime_env["working_dir"]

    return runtime_env


def upload_and_rewrite_workspace_working_dir(
    *,
    api_client: DefaultApi,
    existing_runtime_env: Dict[str, Any],
    workspace_id: str,
    cluster_id: str,
    log: BlockLogger,
) -> Dict[str, Any]:
    """
    Push Workspace working dir to S3 bucket and rewrite the working_dir field with the destination uri

    If the upload_path is not specified by the user, we will get the s3 bucket
    name based on the cloud. We then rewrite the working_dir to the S3 path
    so that the launched service will read from S3 directly.

    The S3 path: s3://{s3_bucket_name}/{org_id}/{cloud_id}/workspace_snapshots/{workspace_id}/{backup_zip}
    """
    decorated_cluster = api_client.get_decorated_cluster_api_v2_decorated_sessions_cluster_id_get(
        cluster_id
    ).result
    cloud_id = decorated_cluster.cloud.id

    s3_bucket_name = _get_cloud_s3_bucket_from_cluster(api_client, decorated_cluster)
    org_id = _get_organization_id(api_client)
    new_runtime_env = copy.deepcopy(existing_runtime_env)
    pre_upload_working_dir = new_runtime_env["working_dir"]

    new_runtime_env[
        "upload_path"
    ] = f"s3://{s3_bucket_name}/{org_id}/{cloud_id}/workspace_snapshots/{workspace_id}"
    new_runtime_env = upload_and_rewrite_working_dir(new_runtime_env)

    log.info(
        f"Uploaded working directory content from {Path(pre_upload_working_dir).absolute()} to {new_runtime_env['working_dir']}"
    )
    return new_runtime_env


def _get_organization_id(api_client: DefaultApi):
    user_info: UserInfo = (api_client.get_user_info_api_v2_userinfo_get().result)
    orgs = user_info.organizations
    return orgs[0].id


def _get_cloud_s3_bucket_from_cluster(
    api_client: DefaultApi, decorated_cluster: DecoratedSession
) -> Optional[str]:
    """
    If the cloud has an associated aws s3 bucket, we return its name.

    Please note that this is only for v2 clouds where customers have their
    own S3 buckets.
    """
    cloud_id = decorated_cluster.cloud.id
    cloud: CloudWithCloudResource = api_client.get_cloud_with_cloud_resource_api_v2_clouds_with_cloud_resource_router_cloud_id_get(
        cloud_id
    ).result
    if cloud.provider == CloudProviders.GCP:
        raise click.ClickException(
            "Currently launching a service from workspaces in a GCP cloud is not supported. "
            "Please contact Anyscale support for more info."
        )
    if cloud and cloud.cloud_resource and cloud.cloud_resource.aws_s3_id:
        s3_bucket_name = cloud.cloud_resource.aws_s3_id
        # Buckets can also be in format of "arn:aws:s3:::bucket/path" instead of "bucket/path"
        if s3_bucket_name.startswith(S3_BUCKET_ARN_PREFIX):
            s3_bucket_name = AwsArn.from_string(s3_bucket_name).resource_id
        return s3_bucket_name
    else:
        return None
